/*
 * This file is part of FGGL.
 *
 * FGGL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 *
 * FGGL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with FGGL.
 * If not, see <https://www.gnu.org/licenses/>.
 */

//
// Created by webpigeon on 02/01/23.
//

#ifndef FGGL_DEMO_INCLUDE_HEXBOARD_CAMERA_HPP
#define FGGL_DEMO_INCLUDE_HEXBOARD_CAMERA_HPP

#include "fggl/math/types.hpp"

namespace demo::hexboard {

	constexpr float TRAUMA_DECAY = 0.01F;
	constexpr float TRAUMA_LARGE = 0.5F;
	constexpr float TRAUMA_SMALL = 0.1F;

	constexpr float SPEED_SLOW = 0.01F;
	constexpr float SPEED_MEDIUM = 0.1F;
	constexpr float SPEED_FAST = 0.5F;

	constexpr fggl::math::vec2 ndc_to_screen(fggl::math::vec2 ndcPos, fggl::math::vec2 screenSize) {
		return fggl::math::vec2(
			fggl::math::rescale_ndc(ndcPos.x, 0, screenSize.x),
			fggl::math::rescale_ndc(ndcPos.y, 0, screenSize.y)
		);
	}

	/**
	 * A 2D 'juiced' camera for grid worlds.
	 *
	 * This camera is based on HexBoard, and the GDC talk around this concept. Many of the ideas implemented here
	 * come from that source.
	 *
	 * @see https://www.youtube.com/watch?v=tu-Qe66AvtY
	 */
	class Camera2D {
		public:
			using Point = fggl::math::vec2;

			/**
			 * Apply active camera effects.
			 *
			 * This method must be called once per frame of the camera's visual effects are to be animated.
			 *
			 * @param delta the amount of time that has passed
			 */
			void update(float delta);

			inline fggl::math::vec2 unproject(fggl::math::vec2 screenPos) const {
				auto location = screenPos;
				location.x += 1920/2.0F;
				location.y += 1080/2.0F;
				return location + m_location;
//				return screenPos + m_location;
			}

			inline fggl::math::vec2 project(fggl::math::vec2 worldPos) const {
				return worldPos - m_location;
			}

			/**
			 * Move the camera to a new location.
			 *
			 * This will apply any movement effects the camera has applied to it. As a result movement to the location
			 * will not be instantaneous.
			 *
			 * @param newTarget the new target location
			 */
			inline void moveTo(Point newTarget) {
				m_target = newTarget;
			}

			/**
			 * Move the camera by a defined amount.
			 *
			 * This will apply any movement effects the camera has applied to it. As a result movement to the location
			 * will not be instantaneous.
			 *
			 * @param delta the amount to add to the target, negative values will subject from the target location
			 */
			inline void moveBy(Point delta) {
				m_target += delta;
			}

			/**
			 * Instantaneously move the camera to a new location.
			 *
			 * @param the new camera location
			 */
			inline void teleportTo(Point newTarget) {
				m_location = newTarget;
				m_target = newTarget;
			}

			/**
			 * Instantaneously adjust the camera location.
			 *
			 * @param delta the amount to adjust the camera location by
			 */
			inline void teleportBy(Point delta) {
				m_location += delta;
				m_target = m_location;
			}

			/**
			 * Get the current view offset of this camera.
			 *
			 * @return the current location this camera is pointing at
			 */
			inline Point getFocusLocation() const {
				auto offset = m_location;
				offset.x += 1920/2.0F;
				offset.y += 1080/2.0F;
				return -offset;
			}

		private:
			Point m_location;
			Point m_target;
			float m_trauma;
			float m_scale = SPEED_MEDIUM;

	};

}

#endif //FGGL_DEMO_INCLUDE_HEXBOARD_CAMERA_HPP
